/*
This demo example starts DDC1 streaming (or DDC1 stream playback), shows IF, DDC1, DDC2 and Audio spectrums 
and plays the final audio through speakers.
You can use + and - keys to tune the DDC1.
You can use * and / keys to tune the DDC2.
*/

#include <stdio.h>
#include <dlfcn.h>
#include <errno.h>
#include <unistd.h>
#include <math.h>
#include <pthread.h>
#include <complex.h>
#include <string.h>
#include <curses.h>			// requires curses library
#include "g35ddcapi.h"
#include <pulse/simple.h>	// requires pulse library
#include <pulse/error.h>
#include <time.h>

//#define PLAYBACK		// use this define to test DDC1 playback. Otherwise the receiver DDC1 streaming is used.
//#define DEMO			// use this define to open a demo receiver. Otherwise a real device will be opened.

pa_simple *s = NULL;	// handle to audio device
// FFT stuff
typedef double complex cplx;
void CreateWindow();
void fft(cplx buf[],bool copyback);
void Translate(cplx fftbuff[],char *outbuff,bool invert,int *min,int *max);

// counters
unsigned long long int TotalAudio=0,TotalDDC1=0,TotalDDC2=0;
unsigned long long int StartTime;

#define FFT_SIZE 64
cplx IF_FFT_Buff[FFT_SIZE];
cplx DDC1_FFT_Buff[FFT_SIZE];
cplx DDC2_FFT_Buff[FFT_SIZE];
cplx Audio_FFT_Buff[FFT_SIZE];

#define DB_PRECISION 8
char dB_Char[DB_PRECISION]={'0','4','5','1','2','6','3','7'};

char BufferIndexArray[4]={'/','-','\\','|'};
int IF_BufferIndex=0,DDC1_BufferIndex=0,DDC2_BufferIndex=0,Audio_BufferIndex=0;

char IF_FFT_Out[6*FFT_SIZE+1];
char DDC1_FFT_Out[6*FFT_SIZE+1];
char DDC2_FFT_Out[6*FFT_SIZE+1];
char Audio_FFT_Out[6*FFT_SIZE+1];

int MinD1=0,MaxD1=0,MinD2=0,MaxD2=0,MinA=0,MaxA=0,MinI=0,MaxI=0;

void IFCallback(const int16_t *Buffer,uint32_t NumberOfSamples,uint16_t MaxADCAmplitude,uint32_t ADCSampleRate,uintptr_t UserData) {
	int i;
	IF_BufferIndex=(IF_BufferIndex+1)&3;
	for (i=0;i<FFT_SIZE;i++) IF_FFT_Buff[i]=(double)Buffer[i]/32768.0;
	fft(IF_FFT_Buff,false);
	Translate(IF_FFT_Buff,IF_FFT_Out,false,&MinI,&MaxI);
	}
void DDC1StreamCallback(const void *Buffer,uint32_t NumberOfSamples,uint32_t BitsPerSample,uintptr_t UserData) {
	int i;
	DDC1_BufferIndex=(DDC1_BufferIndex+1)&3;
	TotalDDC1+=NumberOfSamples;
	for (i=0;i<FFT_SIZE;i++) DDC1_FFT_Buff[i]=(double)((int*)Buffer)[i]/2e9+I*((double)((int*)Buffer)[i+1]/2e9);
	fft(DDC1_FFT_Buff,true);
	Translate(DDC1_FFT_Buff,DDC1_FFT_Out,true,&MinD1,&MaxD1);
	}
int DDC1PlaybackStreamCallback(void *Buffer,uint32_t NumberOfSamples,uint32_t BitsPerSample,uintptr_t UserData) {
#ifdef PLAYBACK
	unsigned int i;
	int *buff=(int*)Buffer;
	for (i=0;i<NumberOfSamples;i++) {
		double angle=2*M_PI*10*i/NumberOfSamples;
		buff[i*2]=(int)(2e7*cos(angle));
		buff[i*2+1]=(int)(-2e7*sin(angle));
		}
#endif
	return 1;
	}
void DDC2StreamCallback(uint32_t Channel,const float *Buffer,uint32_t NumberOfSamples,uintptr_t UserData) {
	int i;
	DDC2_BufferIndex=(DDC2_BufferIndex+1)&3;
	TotalDDC2+=NumberOfSamples;
	for (i=0;i<FFT_SIZE;i++) DDC2_FFT_Buff[i]=Buffer[i]+I*(Buffer[i+1]);
	fft(DDC2_FFT_Buff,true);
	Translate(DDC2_FFT_Buff,DDC2_FFT_Out,true,&MinD2,&MaxD2);
	}
void DDC2PreprocessedStreamCallback(uint32_t Channel,const float *Buffer,uint32_t NumberOfSamples,float SlevelPeak,float SlevelRMS,uintptr_t UserData) {}
void AudioStreamCallback(uint32_t Channel,uint32_t Type,const float *Buffer,uint32_t NumberOfSamples,uintptr_t UserData) {
	if (!Channel&&!Type) {		// first demodulator (RX) and 
		int i;
		TotalAudio+=NumberOfSamples;
		Audio_BufferIndex=(Audio_BufferIndex+1)&3;
		pa_simple_write(s,Buffer,(size_t)NumberOfSamples*sizeof(float),NULL);
		for (i=0;i<FFT_SIZE;i++) Audio_FFT_Buff[i]=Buffer[i];//+I*(Buffer[i+1]);
		fft(Audio_FFT_Buff,false);
		Translate(Audio_FFT_Buff,Audio_FFT_Out,false,&MinA,&MaxA);
		}
	}
int AudioPlaybackStreamCallback(uint32_t Channel,float *Buffer,uint32_t NumberOfSamples,uintptr_t UserData) {return 0;}

int main(void)
{
G35DDC_OPEN_DEVICE OpenDevice;
G35DDC_CLOSE_DEVICE CloseDevice;
G35DDC_START_IF StartIF;
G35DDC_START_DDC1 StartDDC1;
G35DDC_STOP_IF StopIF;
G35DDC_SET_POWER SetPower;
G35DDC_SET_DDC1 SetDDC1;
G35DDC_STOP_DDC1 StopDDC1;
G35DDC_START_DDC2 StartDDC2;
G35DDC_STOP_DDC2 StopDDC2;
G35DDC_START_AUDIO StartAudio;
G35DDC_STOP_AUDIO StopAudio;
G35DDC_SET_DDC1_FREQUENCY SetDDC1Frequency;
G35DDC_SET_DDC2_FREQUENCY SetDDC2Frequency;
G35DDC_START_DDC1_PLAYBACK StartDDC1Playback;
G35DDC_GET_DDC1 GetDDC1;
G35DDC_GET_DDC2 GetDDC2;
G35DDC_SET_DEMODULATOR_FILTER_BANDWIDTH SetDemodulatorFilterBandwidth;
G35DDC_SET_DEMODULATOR_MODE SetDemodulatorMode;
G35DDC_SET_AGC SetAGC;
G35DDC_SET_AGC_PARAMS SetAGCParams;
G35DDC_GET_DEVICE_INFO GetDeviceInfo;
G35DDC_SET_CALLBACKS SetCallbacks;
G35DDC_CALLBACKS Callbacks;

Callbacks.IFCallback=IFCallback;
Callbacks.DDC1StreamCallback=DDC1StreamCallback;
Callbacks.DDC1PlaybackStreamCallback=DDC1PlaybackStreamCallback;
Callbacks.DDC2StreamCallback=DDC2StreamCallback;
Callbacks.DDC2PreprocessedStreamCallback=DDC2PreprocessedStreamCallback;
Callbacks.AudioStreamCallback=AudioStreamCallback;
Callbacks.AudioPlaybackStreamCallback=AudioPlaybackStreamCallback;

int32_t hDevice=0;
void *API;

CreateWindow();
int Freq1=15000000,Freq2=0;
int error=0;
struct timespec tt;
int i;

// open the audio output
pa_sample_spec ss;
ss.format=PA_SAMPLE_FLOAT32LE,
ss.rate=48000;
ss.channels=1;
if (!(s=pa_simple_new(NULL,"g35ddcdemo",PA_STREAM_PLAYBACK,NULL,"playback",&ss,NULL,NULL,&error))) fprintf(stderr, __FILE__": pa_simple_new() failed: %s\n", pa_strerror(error));

// prepare the buffers for color spectrums
for (i=0;i<FFT_SIZE;i++) {
	IF_FFT_Out[6*i]=DDC1_FFT_Out[6*i]=DDC2_FFT_Out[6*i]=Audio_FFT_Out[6*i]=27;
	IF_FFT_Out[6*i+1]=DDC1_FFT_Out[6*i+1]=DDC2_FFT_Out[6*i+1]=Audio_FFT_Out[6*i+1]='[';
	IF_FFT_Out[6*i+2]=DDC1_FFT_Out[6*i+2]=DDC2_FFT_Out[6*i+2]=Audio_FFT_Out[6*i+2]='4';
	IF_FFT_Out[6*i+3]=DDC1_FFT_Out[6*i+3]=DDC2_FFT_Out[6*i+3]=Audio_FFT_Out[6*i+3]='0';
	IF_FFT_Out[6*i+4]=DDC1_FFT_Out[6*i+4]=DDC2_FFT_Out[6*i+4]=Audio_FFT_Out[6*i+4]='m';	
	IF_FFT_Out[6*i+5]=DDC1_FFT_Out[6*i+5]=DDC2_FFT_Out[6*i+5]=Audio_FFT_Out[6*i+5]=' ';	
	}
IF_FFT_Out[6*FFT_SIZE/2]=Audio_FFT_Out[6*FFT_SIZE/2]=0;
DDC1_FFT_Out[6*FFT_SIZE]=DDC2_FFT_Out[6*FFT_SIZE]=0;

API=dlopen("libg35ddcapi.so",RTLD_LAZY);
if (API) {
	OpenDevice=(G35DDC_OPEN_DEVICE)dlsym(API,"OpenDevice");
	CloseDevice=(G35DDC_CLOSE_DEVICE)dlsym(API,"CloseDevice");
	SetPower=(G35DDC_SET_POWER)dlsym(API,"SetPower");
	SetDDC1=(G35DDC_SET_DDC1)dlsym(API,"SetDDC1");
	GetDDC2=(G35DDC_GET_DDC2)dlsym(API,"GetDDC2");
	GetDDC1=(G35DDC_GET_DDC1)dlsym(API,"GetDDC1");
	StartIF=(G35DDC_START_IF)dlsym(API,"StartIF");
	StopIF=(G35DDC_STOP_IF)dlsym(API,"StopIF");
	StopAudio=(G35DDC_STOP_AUDIO)dlsym(API,"StopAudio");
	StartDDC1=(G35DDC_START_DDC1)dlsym(API,"StartDDC1");
	StopDDC1=(G35DDC_STOP_DDC1)dlsym(API,"StopDDC1");
	StartDDC2=(G35DDC_START_DDC2)dlsym(API,"StartDDC2");
	StartAudio=(G35DDC_START_AUDIO)dlsym(API,"StartAudio");
	StopDDC2=(G35DDC_STOP_DDC2)dlsym(API,"StopDDC2");
	SetDDC2Frequency=(G35DDC_SET_DDC2_FREQUENCY)dlsym(API,"SetDDC2Frequency");
	SetDDC1Frequency=(G35DDC_SET_DDC1_FREQUENCY)dlsym(API,"SetDDC1Frequency");
	StartDDC1Playback=(G35DDC_START_DDC1_PLAYBACK)dlsym(API,"StartDDC1Playback");
	SetDemodulatorFilterBandwidth=(G35DDC_SET_DEMODULATOR_FILTER_BANDWIDTH)dlsym(API,"SetDemodulatorFilterBandwidth");
	SetDemodulatorMode=(G35DDC_SET_DEMODULATOR_MODE)dlsym(API,"SetDemodulatorMode");
	SetAGC=(G35DDC_SET_AGC)dlsym(API,"SetAGC");
	SetAGCParams=(G35DDC_SET_AGC_PARAMS)dlsym(API,"SetAGCParams");
	SetCallbacks=(G35DDC_SET_CALLBACKS)dlsym(API,"SetCallbacks");
	GetDeviceInfo=(G35DDC_GET_DEVICE_INFO)dlsym(API,"GetDeviceInfo");
//	=()dlsym(API,"");

#ifdef DEMO
	hDevice=OpenDevice(G3XDDC_OPEN_DEMO);
#else
	hDevice=OpenDevice(G3XDDC_OPEN_FIRST);
#endif
	if (hDevice) {
		G35DDC_DEVICE_INFO info;
		G3XDDC_DDC_INFO info1={0},info2={0};
		GetDeviceInfo(hDevice,&info,sizeof(info));
		printf("Device %8s opened, handle=%d\n",info.SerialNumber,hDevice);

		SetCallbacks(hDevice,&Callbacks,(uintptr_t)NULL);
		int ret=SetPower(hDevice,1); printf("SetPower, ret=%d, errno=%d\n",ret,errno);
		ret=StartIF(hDevice,300); printf("StartIF, ret=%d, errno=%d\n",ret,errno);
		ret=SetDDC1(hDevice,22); printf("SetDDC1, ret=%d, errno=%d\n",ret,errno);
		ret=GetDDC1(hDevice,NULL,&info1); printf("GetDDC1, ret=%d, errno=%d\n",ret,errno);
#ifdef PLAYBACK
		ret=StartDDC1Playback(hDevice,30000,32); printf("StartDDC1Playback, ret=%d, errno=%d\n",ret,errno);
#else
		ret=StartDDC1(hDevice,50000); printf("StartDDC1, ret=%d, errno=%d\n",ret,errno);
		ret=SetDDC1Frequency(hDevice,Freq1); printf("SetDDC1Freq, ret=%d, errno=%d\n",ret,errno);
#endif
		clock_gettime(CLOCK_REALTIME,&tt);
		StartTime=tt.tv_sec;
		StartTime*=1000000000;
		StartTime+=tt.tv_nsec;

		ret=StartDDC2(hDevice,0,40000); printf("StartDDC2, ret=%d, errno=%d\n",ret,errno);
		GetDDC2(hDevice,NULL,&info2);
		ret=SetDDC2Frequency(hDevice,0,Freq2); printf("SetDDC2Freq, ret=%d, errno=%d\n",ret,errno);
		SetDemodulatorMode(hDevice,0,G3XDDC_MODE_CW);
		ret=StartAudio(hDevice,0,2000); printf("StartAudio, ret=%d, errno=%d\n",ret,errno);
		SetDemodulatorFilterBandwidth(hDevice,0,25000);
		SetAGC(hDevice,0,TRUE);
		SetAGCParams(hDevice,0,0.1,0.1,-20);
		puts("Streaming started, press a key");
		getchar();
		printf("\E[H\E[2J");
		WINDOW *w=initscr();
		cbreak();
		noecho();
		nodelay(w,TRUE);
			
		bool end=false;
		do {
			printf("\033[6;0HDDC1: %dkHz/%dkHz, Freq=%dkHz\t\tDDC2=%dkHz/%dkHz, Freq=%dkHz     \n",info1.SampleRate/1000,info1.Bandwidth/1000,Freq1/1000,info2.SampleRate/1000,info2.Bandwidth/1000,Freq2/1000);
			printf("\033[8;0HIF:     %s\x1b[0m%c %d+%d  ",IF_FFT_Out,(unsigned int)BufferIndexArray[IF_BufferIndex],MinI,MaxI);
			printf("\033[10;0HDDC1: %s\x1b[0m%c %d+%d  ",DDC1_FFT_Out,BufferIndexArray[DDC1_BufferIndex],MinD1,MaxD1);
			printf("\033[12;0HDDC2: %s\x1b[0m%c %d+%d  ",DDC2_FFT_Out,BufferIndexArray[DDC2_BufferIndex],MinD2,MaxD2);
			printf("\033[14;0HAudio: %s\x1b[0m%c %d+%d  ",Audio_FFT_Out,BufferIndexArray[Audio_BufferIndex],MinA,MaxA);
			usleep(10000);
			int c=getch();
			if (c==27) end=true;
			else if (c=='+') {Freq1+=1000;SetDDC1Frequency(hDevice,Freq1);}
			else if (c=='-') {Freq1-=1000;SetDDC1Frequency(hDevice,Freq1);}
			else if (c=='/') {Freq2+=1000;SetDDC2Frequency(hDevice,0,Freq2);}
			else if (c=='*') {Freq2-=1000;SetDDC2Frequency(hDevice,0,Freq2);}
			} while (!end);
		clock_gettime(CLOCK_REALTIME,&tt);
		unsigned long long int t=tt.tv_sec;
		t*=1000000000;
		t+=tt.tv_nsec;
		t-=StartTime;
		t/=1000000;

		endwin();

		printf("DDC1 rate=%d kHz, DDC2 rate=%dkHz, Audio rate=%d kHz\n\n",(unsigned int)(TotalDDC1/t),(unsigned int)(TotalDDC2/t),(unsigned int)(TotalAudio/t));

		puts("Stopping IF");
		StopIF(hDevice);
		puts("Stopping Audio");
		StopAudio(hDevice,0);
		puts("Stopping DDC2");
		StopDDC2(hDevice,0);
		puts("Stopping DDC1");
		StopDDC1(hDevice);
		puts("Closing the device");
		CloseDevice(hDevice);
		puts("Device closed");		
		}
	else printf("Failed to open device. Error code=%d\n",errno);
	puts("Closing the API");
	dlclose(API);
    }
else printf("Failed to load libg35ddcapi.so. %s\n",dlerror());

if (s) pa_simple_free(s);
puts("Closing the demo");
return 0;
}

//---------------------------------------------------------------------------------------------------
double window[FFT_SIZE];
void CreateWindow()
{
double sum=0.0;
int i;
for (i=0;i<FFT_SIZE;i++) sum+=(window[i]=0.42-0.5*cos(2.0*M_PI*i/(double)(FFT_SIZE-1))+0.08*cos(4.0*M_PI*i/(double)(FFT_SIZE-1)));
sum=1.0/sum;
for (i=0;i<FFT_SIZE;i++) window[i]*=sum;
}
//--------------------------------------------------------------------------------
void Translate(cplx fftbuff[],char *outbuff,bool invert,int *min,int *max)
{
double maxdB=-100000,mindB=0;
int i;
for (i=0;i<FFT_SIZE;i++) {
	double a=10*log10(creal(fftbuff[i])*creal(fftbuff[i])+cimag(fftbuff[i])*cimag(fftbuff[i])+1e-20);
	fftbuff[i]=a;
	if (a>maxdB) maxdB=a;
	if (a<mindB) mindB=a;
	}
maxdB-=mindB;
if (invert) for (i=0;i<FFT_SIZE/2;i++) {
	unsigned int index=(unsigned int)((DB_PRECISION-1)*((creal(fftbuff[i])-mindB)/maxdB));
	outbuff[6*(FFT_SIZE/2-1-i)+3]=dB_Char[index];
	index=(unsigned int)((DB_PRECISION-1)*((creal(fftbuff[FFT_SIZE/2+i])-mindB)/maxdB));
	outbuff[6*(FFT_SIZE-1-i)+3]=dB_Char[index];
	}
else for (i=0;i<FFT_SIZE;i++) {
	unsigned int index=(unsigned int)((DB_PRECISION-1)*((creal(fftbuff[i])-mindB)/maxdB));
	if (index>=DB_PRECISION) index=DB_PRECISION-1;
	outbuff[6*i+3]=dB_Char[index];
	}
*min=(int)mindB;
*max=(int)maxdB;
}
//--------------------------------------------------------------------------------
void _fft(cplx buf[], cplx out[],int step)
{
int i;
if (step<FFT_SIZE) {
	_fft(out,buf,step*2);
	_fft(out+step,buf+step,step*2);

	for (i=0;i<FFT_SIZE;i+=2*step) {
		cplx t=cexp(-1*I*M_PI*i/FFT_SIZE)*out[i+step];
		buf[i/2]=out[i]+t;
		buf[(i+FFT_SIZE)/2]=out[i]-t;
		}
	}
}
void fft(cplx buf[],bool copyback)
{
int i;
cplx out[FFT_SIZE];
for (i=0;i<FFT_SIZE;i++) out[i] = buf[i]*window[i];
_fft(buf,out,1);
if (copyback) for (i=0;i<FFT_SIZE;i++) buf[i]=out[i];
}
//--------------------------------------------------------------------------------
